import os
import sys
from traceback import print_exc

import pytest

from astropy.coordinates import SkyCoord
from astropy import units as u

from qtpy.QtWebEngineWidgets import WEBENGINE

from matplotlib.testing.compare import compare_images

M42 = SkyCoord.from_name('M42')

DATA = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))


def check_silent_output(capsys):
    out, err = capsys.readouterr()
    assert out.strip() == ""
    assert err.strip() == ""


class TestWWTWidget:

    def test_settings(self, capsys, wwt_qt_client):
        wwt_qt_client.constellation_figures = True
        wwt_qt_client.constellation_figures = False
        wwt_qt_client.wait(1)
        check_silent_output(capsys)

    def test_methods(self, capsys, wwt_qt_client):
        wwt_qt_client.center_on_coordinates(M42, fov=10 * u.deg)
        wwt_qt_client.wait(1)
        check_silent_output(capsys)

    def test_coordinates(self, capsys, wwt_qt_client):
        wwt_qt_client.center_on_coordinates(M42, fov=10 * u.deg)
        assert M42.separation(wwt_qt_client.get_center()).arcsec < 1.e-6
        wwt_qt_client.wait(1)
        check_silent_output(capsys)

    def test_annotations(self, capsys, wwt_qt_client):
        circle = wwt_qt_client.add_circle()
        circle.opacity = 0.8
        circle.set_center(M42)
        wwt_qt_client.wait(1)
        check_silent_output(capsys)


# The following is a template for a script that will allow developers who see
# a failure in CI to re-create the files that were generated in the
# continuous integration easily.
REPRODUCABILITY_SCRIPT = """
################################################################################
# Export the images that were generated in the continuous integration for pywwt.
# Just copy and paste all the code between here and '# End of script' into a
# local file and run it with Python. You can then check if the differences
# make sense, and if so, update the expected images.

import base64

expected = base64.b64decode('{expected}')

with open('expected.png', 'wb') as f:
    f.write(expected)

actual = base64.b64decode('{actual}')

with open('actual.png', 'wb') as f:
    f.write(actual)

# End of script
################################################################################
"""


def assert_widget_image(tmpdir, widget, filename):
    actual = tmpdir.join(filename).strpath
    widget.render(actual)
    framework = 'webengine' if WEBENGINE else 'webkit'
    if sys.platform.startswith('win') and not WEBENGINE:
        framework += '_win'
    elif sys.platform.startswith('darwin'):
        framework += '_osx'
    expected = os.path.join(DATA, framework, filename)
    try:
        msg = compare_images(expected, actual, tol=1.5)
    except Exception:
        msg = 'Image comparison failed:'
        print_exc()

    if msg is not None:

        from base64 import b64encode

        with open(expected, 'rb') as f:
            expected = b64encode(f.read()).decode()

        with open(actual, 'rb') as f:
            actual = b64encode(f.read()).decode()

        if os.environ.get('CI', 'false').lower() == 'true':
            print(REPRODUCABILITY_SCRIPT.format(actual=actual, expected=expected))

        pytest.fail(msg, pytrace=False)


def test_full(tmpdir, capsys, wwt_qt_client):

    # Test a whole session, with image comparison along the way.

    wwt = wwt_qt_client
    wwt.foreground_opacity = 1.

    # The crosshairs are currently broken on Mac/Linux but work on Windows.
    # For consistency, we turn it off here so that the results are the same
    # on all platforms.
    wwt.crosshairs = False

    wwt.wait(4)

    assert_widget_image(tmpdir, wwt, 'test_full_step0.png')

    gc = SkyCoord(0, 0, unit=('deg', 'deg'), frame='galactic')
    wwt.center_on_coordinates(gc, 60 * u.deg)

    wwt.wait(4)

    assert_widget_image(tmpdir, wwt, 'test_full_step1.png')

    wwt.constellation_boundary_color = 'red'
    wwt.constellation_figure_color = 'green'
    wwt.constellation_selection_color = 'blue'

    wwt.constellation_boundaries = True
    wwt.constellation_figures = True

    wwt.wait(4)

    assert_widget_image(tmpdir, wwt, 'test_full_step2.png')

    wwt.constellation_selection = True

    wwt.crosshairs = False
    wwt.ecliptic = True
    wwt.grid = True

    wwt.wait(4)

    assert_widget_image(tmpdir, wwt, 'test_full_step3.png')

    wwt.foreground = 'SFD Dust Map (Infrared)'

    wwt.wait(4)

    assert_widget_image(tmpdir, wwt, 'test_full_step4.png')

    wwt.foreground = "Black Sky Background"
    wwt.background = "Black Sky Background"
    wwt.foreground_opacity = 0

    wwt.center_on_coordinates(gc, 30 * u.deg)

    coord = SkyCoord(5, 0.5, unit=('deg', 'deg'), frame='galactic')

    circle1 = wwt.add_circle()
    circle1.set_center(coord)
    circle1.radius = 10 * u.pixel
    circle1.line_width = 5 * u.pixel
    circle1.line_color = 'green'
    circle1.fill = False
    circle1.opacity = 0.5

    coord = SkyCoord(-5, -0.5, unit=('deg', 'deg'), frame='galactic')

    circle2 = wwt.add_circle()
    circle2.set_center(coord)
    circle2.radius = 2 * u.degree
    circle2.line_width = 5 * u.pixel
    circle2.line_color = 'green'
    circle2.fill = True
    circle2.fill_color = 'orange'
    circle2.opacity = 1

    coord = SkyCoord([1, 4, 0], [0, 0, -5], unit=('deg', 'deg'), frame='galactic')

    poly = wwt.add_polygon()
    poly.add_point(coord[0])
    poly.add_point(coord[1])
    poly.add_point(coord[2])
    poly.fill = True
    poly.line_color = 'red'
    poly.fill_color = 'yellow'
    poly.line_width = 2 * u.pixel

    coord = SkyCoord([10, 5, 2], [5, 2, 2], unit=('deg', 'deg'), frame='galactic')

    polyline = wwt.add_line()
    polyline.add_point(coord[0])
    polyline.add_point(coord[1])
    polyline.add_point(coord[2])
    polyline.color = 'green'
    polyline.width = 3 * u.pixel

    wwt.wait(4)

    assert_widget_image(tmpdir, wwt, 'test_full_step5.png')



import csv
import sys
import codecs
import struct
from datetime import datetime, timedelta

import numpy as np
import matplotlib.cm as mcm
from matplotlib.colors import LogNorm, Normalize
from astropy.utils.console import ProgressBar
from dateutil import tz, parser

__all__ = ['map_array_to_colors', 'generate_utc_times',
           'convert_xyz_to_spherical', 'write_data_to_csv']


def map_array_to_colors(arr, cmap, scale="linear",
                        vmin=None, vmax=None):
    """
    Map a NumPy array to a colormap using Matplotlib.

    Parameters
    ----------
    array : `numpy.ndarray`
        The array to be mapped onto a colormap.
    cmap : `str`
        The name of a Matplotlib colormap.
    scale : {'linear' | 'log'}, optional
        The scaling of the data as mapped to the colormap.
    vmin : `float`, optional
        The value associated with the low end of the colormap. By default this
        is the minimum value of ``arr``.
    vmax : `float`, optional
        The value associated with the high end of the colormap. By default this
        is the maximum value of ``arr``.

    Returns
    -------
    colors : `list`
        A list of colors, encoded in ARGB hex values.
    """
    if vmin is None:
        vmin = arr.min()
    if vmax is None:
        vmax = arr.max()

    if scale == "linear":
        norm = Normalize(vmin=vmin, vmax=vmax)
    elif scale == "log":
        norm = LogNorm(vmin=vmin, vmax=vmax)

    my_cmap = mcm.ScalarMappable(norm=norm, cmap=cmap)
    colors = my_cmap.to_rgba(arr, bytes=True)

    hex_colors = []

    for color in colors:
        hex_color = codecs.getencoder('hex')(struct.pack('BBB', *color[:3]))[0]
        hex_colors.append("FF" + hex_color.decode().upper())

    return hex_colors


def generate_utc_times(num_steps, step_size, start_time=None):
    """
    Generate a series of equally linearly spaced times in UTC.

    Parameters
    ----------
    num_steps : `int`
        The number of times to generate.
    step_size : `dict`
        A dictionary corresponding to the step size between the times, with keys
        referring to the unit (seconds, minutes, days, etc.) and the values of
        the step in that unit.
    start_date : `str`, optional
        A string corresponding to a time in the local time zone at which to
        start the timesteps. Default: the current system time. Acceptable
        formats (month/day/year):
        ``"1/1/2010 11:00:00 PM"``,
        ``"1/1/2010 11:30 AM"``,
        ``"1/1/2010 11 am"``,
        ``"1/1/2000"``,
        ``"1/2000"``,

    Returns
    -------
    times : `list`
        A list of formatted date/time strings in UTC readable by WorldWide
        Telescope.
    """
    if start_time is None:
        start_time = datetime.utcnow()
    else:
        start_time = parser.parse(start_time)
        utc_zone = tz.tzutc()
        local_zone = tz.tzlocal()
        # Tell the datetime object that it's in local time zone since
        # datetime objects are 'naive' by default
        local_time = start_time.replace(tzinfo=local_zone)
        # Convert time to UTC
        start_time = local_time.astimezone(utc_zone)

    time_arr = []
    new_time = start_time
    for i in range(num_steps):
        time_arr.append(new_time.strftime("%m/%d/%Y %I:%M:%S %p"))
        new_time += timedelta(**step_size)

    return time_arr


def convert_xyz_to_spherical(x, y, z, is_astro=True, ra_units="degrees"):
    """
    Convert rectangular coordinates (x,y,z) to spherical coordinates
    (Lat, Lon, Alt) or (RA, Dec, Alt).

    Parameters
    ----------
    x : `numpy.ndarray`
        The x-coordinates of the data.
    y : `numpy.ndarray`
        The y-coordinates of the data.
    z : `numpy.ndarray`
        The z-coordinates of the data.
    is_astro : `bool`, optional
        Whether the coordinate system is astronomical (RA, Dec) or geographical
        (Lat, Lon)?
    ra_units : {'degrees', 'hours'}, optional
        The unit of the RA/Lon coordinate.

    Returns
    -------
    spherical : `dict`
        A dict of NumPy arrays corresponding to the positions in spherical
        coordinates.
    """
    if ra_units == "degrees" or not is_astro:
        ra_scale = 1.
    elif ra_units == "hours":
        ra_scale = 24. / 360.
    if is_astro:
        ra_name = "RA"
        dec_name = "DEC"
    else:
        ra_name = "LON"
        dec_name = "LAT"
    coords = {}
    coords["ALT"] = np.sqrt(x * x + y * y + z * z)
    coords[ra_name] = (np.rad2deg(np.arctan2(y, x)) + 180.) * ra_scale
    coords[dec_name] = np.rad2deg(np.arccos(z / coords["ALT"])) - 90.
    return coords


def write_data_to_csv(data, filename, mode="new"):
    """
    Write a dataset to a CSV-formatted file with a data header.

    Parameters
    ----------
    data : `dict`
        The data to be written as a dictionary of NumPy arrays
    filename : `str`
        The filename to write the data to.
    mode : {'new', 'append'}, optional
        Write a "new" file or "append" to an existing file?
    """
    if mode == "new":
        fmode = "w"
    elif mode == "append":
        fmode = "a+"
    if sys.version_info >= (3, 0, 0):
        f = open(filename, fmode, newline='')
    else:
        f = open(filename, fmode + 'b')
    w = csv.DictWriter(f, list(data.keys()))
    if mode == "new":
        w.writeheader()
    num_points = len(list(data.values())[0])
    for i in ProgressBar(list(range(num_points))):
        row = dict([(k, v[i]) for k, v in list(data.items())])
        w.writerow(row)
    f.close()



# This file contains the defintion of the Python part of the WWT Jupyter
# widget. Note that we don't tag each trait from BaseWWTWidget as sync=True
# because we instead use JSON messages to transmit any changes between the
# Python and Javascript parts so that we can re-use this for the Qt client.

import sys
PY2 = sys.version_info[0] == 2

import ipywidgets as widgets
from traitlets import Unicode, Float, default, link

if not PY2:
    from ipyevents import Event as DOMListener

from .core import BaseWWTWidget
from .jupyter_server import serve_file

__all__ = ['WWTJupyterWidget']

if not PY2:
    dom_listener = DOMListener()


@widgets.register
class WWTJupyterWidget(widgets.DOMWidget, BaseWWTWidget):
    """
    A WorldWide Telescope Jupyter widget.
    """

    _view_name = Unicode('WWTView').tag(sync=True)
    _model_name = Unicode('WWTModel').tag(sync=True)
    _view_module = Unicode('pywwt').tag(sync=True)
    _model_module = Unicode('pywwt').tag(sync=True)
    _view_module_version = Unicode('0.6.0').tag(sync=True)
    _model_module_version = Unicode('0.6.0').tag(sync=True)
    _ra = Float(0.0).tag(sync=True)
    _dec = Float(0.0).tag(sync=True)
    _fov = Float(60.0).tag(sync=True)

    def __init__(self):
        widgets.DOMWidget.__init__(self)
        BaseWWTWidget.__init__(self)
        if not PY2:
            dom_listener.source = self
            dom_listener.prevent_default_action = True
            dom_listener.watched_events = ['wheel']
        self._controls = None

    @default('layout')
    def _default_layout(self):
        return widgets.Layout(height='400px', align_self='stretch')

    def _send_msg(self, **kwargs):
        self.send(kwargs)

    def _serve_file(self, filename, extension=''):
        return serve_file(filename, extension=extension)

    def _get_view_data(self, field):
        if field == 'ra':
            return self._ra
        elif field == 'dec':
            return self._dec
        elif field == 'fov':
            return self._fov
        else:
            raise ValueError("'field' should be one of: 'ra', 'dec', or 'fov'")

    @property
    def layer_controls(self):
        if self._controls is None:
            opacity_slider = widgets.FloatSlider(value=self.foreground_opacity,
                                                 min=0, max=1, readout=False)
            foreground_menu = widgets.Dropdown(options=self.available_layers,
                                               value=self.foreground)
            background_menu = widgets.Dropdown(options=self.available_layers,
                                               value=self.background)
            link((opacity_slider, 'value'), (self, 'foreground_opacity'))
            link((foreground_menu, 'value'), (self, 'foreground'))
            link((background_menu, 'value'), (self, 'background'))
            self._controls = widgets.HBox([background_menu, opacity_slider, foreground_menu])
        return self._controls



import uuid
from traitlets import HasTraits, TraitError, validate
from astropy import units as u
from astropy.coordinates import concatenate, SkyCoord
import requests
import numpy as np

from .traits import (Color, ColorWithOpacity, Bool,
                     Float, Unicode, AstropyQuantity)

# The WWT web control API is described here:
# https://worldwidetelescope.gitbook.io/html5-control-reference/

__all__ = ['Annotation', 'Circle', 'Polygon', 'Line', 'FieldOfView']


class Annotation(HasTraits):
    """
    Base class for annotations which provides settings common to all shapes.
    """

    shape = None
    """
    To be specified in the individual shape classes.
    """

    label = Unicode('', help='Contains descriptive text '
                             'for the annotation (`str`)').tag(wwt='label')

    opacity = Float(1, help='Specifies the opacity to be applied to the '
                            'complete annotation (`int`)').tag(wwt='opacity')

    hover_label = Bool(False,
                       help='Specifies whether to render the label '
                            'if the mouse is hovering over the annotation '
                            '(`bool`)').tag(wwt='showHoverLabel')

    tag = Unicode(help='Contains a string for use by the web client '
                       '(`str`)').tag(wwt='tag')

    def __init__(self, parent=None, **kwargs):
        self.parent = parent
        self.observe(self._on_trait_change, type='change')
        self.id = str(uuid.uuid4())
        if all(key in self.trait_names() for key in kwargs):
            self.parent._send_msg(event='annotation_create',
                                  id=self.id, shape=self.shape)
            super(Annotation, self).__init__(**kwargs)
        else:
            raise KeyError('a key doesn\'t match any annotation trait name')

    def _on_trait_change(self, changed):
        # This method gets called anytime a trait gets changed. Since this class
        # gets inherited by the Jupyter widgets class which adds some traits of
        # its own, we only want to react to changes in traits that have the wwt
        # metadata attribute (which indicates the name of the corresponding WWT
        # setting).
        wwt_name = self.trait_metadata(changed['name'], 'wwt')
        if wwt_name is not None:
            self.parent._send_msg(event='annotation_set',
                                  id=self.id,
                                  setting=wwt_name,
                                  value=changed['new'])


class Circle(Annotation):
    """
    A circular annotation.
    """

    shape = 'circle'
    """
    The name of the shape (:class:`str`).
    """

    fill = Bool(False, help='Whether or not the circle should be filled '
                            '(`bool`)').tag(wwt='fill')
    fill_color = ColorWithOpacity('white',
                                  help='Assigns fill color for the circle '
                                       '(`str` or `tuple`)').tag(wwt='fillColor')
    line_color = Color('white', help='Assigns line color for the circle '
                                     '(`str` or `tuple`)').tag(wwt='lineColor')
    line_width = AstropyQuantity(1*u.pixel,
                                 help='Assigns line width in pixels '
                                      '(:class:`~astropy.units.Quantity`)').tag(wwt='lineWidth')
    radius = AstropyQuantity(80*u.pixel,
                             help='Sets the radius for the circle '
                                  '(:class:`~astropy.units.Quantity`)').tag(wwt='radius')

    @validate('line_width')
    def _validate_linewidth(self, proposal):
        if proposal['value'].unit.is_equivalent(u.pixel):
            return proposal['value'].to(u.pixel)
        else:
            raise TraitError('line width must be in pixel equivalent unit')

    @validate('radius')
    def _validate_radius(self, proposal):
        if proposal['value'].unit.is_equivalent(u.pixel):
            return proposal['value'].to(u.pixel)
        elif proposal['value'].unit.is_equivalent(u.degree):
            return proposal['value'].to(u.degree)
        else:
            raise TraitError('radius must be in pixel or degree equivalent unit')

    def set_center(self, coord):
        """
        Set the center coordinates of the circle object.

        Parameters
        ----------
        coord : `~astropy.units.Quantity`
            The coordinates of the desired center of the circle.
        """
        coord_icrs = coord.icrs
        self.parent._send_msg(event='circle_set_center', id=self.id,
                              ra=coord_icrs.ra.degree,
                              dec=coord_icrs.dec.degree)

    def remove(self):
        """
        Removes the specified annotation from the current view.
        """
        self.parent._send_msg(event='remove_annotation', id=self.id)

    def _on_trait_change(self, changed):
        if changed['name'] == 'radius':
            if changed['new'].unit.is_equivalent(u.pixel):
                self.parent._send_msg(event='annotation_set', id=self.id,
                                      setting='skyRelative', value=False)
            elif changed['new'].unit.is_equivalent(u.degree):
                self.parent._send_msg(event='annotation_set', id=self.id,
                                      setting='skyRelative', value=True)
        if isinstance(changed['new'], u.Quantity):
            changed['new'] = changed['new'].value

        super(Circle, self)._on_trait_change(changed)


class Polygon(Annotation):
    """
    A polygon annotation.
    """

    shape = 'polygon'
    """
    The name of the shape (:class:`str`).
    """

    fill = Bool(False, help='Whether or not the polygon should be filled '
                            '(`bool`)').tag(wwt='fill')
    fill_color = ColorWithOpacity('white',
                                  help='Assigns fill color for the polygon '
                                       '(`str` or `tuple`)').tag(wwt='fillColor')
    line_color = Color('white', help='Assigns line color for the polygon '
                                     '(`str` or `tuple`)').tag(wwt='lineColor')
    line_width = AstropyQuantity(1*u.pixel,
                                 help='Assigns line width in pixels '
                                      '(:class:`~astropy.units.Quantity`)').tag(wwt='lineWidth')

    @validate('line_width')
    def _validate_linewidth(self, proposal):
        if proposal['value'].unit.is_equivalent(u.pixel):
            return proposal['value'].to(u.pixel)
        else:
            raise TraitError('line width must be in pixel equivalent unit')

    def add_point(self, coord):
        """
        Add one or more points to a polygon object.

        If you want to fill the polygon, you should ensure that the vertices
        form a counter-clockwise polygon.

        Parameters
        ----------
        coord : `~astropy.units.Quantity`
            The coordinates of the desired point(s).
        """
        coord_icrs = coord.icrs
        if coord_icrs.isscalar: # if coord only has one point
            self.parent._send_msg(event='polygon_add_point', id=self.id,
                                  ra=coord_icrs.ra.degree,
                                  dec=coord_icrs.dec.degree)
        else:
            for point in coord_icrs:
                self.parent._send_msg(event='polygon_add_point', id=self.id,
                                      ra=point.ra.degree,
                                      dec=point.dec.degree)

    def remove(self):
        """
        Removes the specified annotation from the current view.
        """
        self.parent._send_msg(event='remove_annotation', id=self.id)

    def _on_trait_change(self, changed):
        if isinstance(changed['new'], u.Quantity):
            changed['new'] = changed['new'].value

        super(Polygon, self)._on_trait_change(changed)


class Line(Annotation):
    """
    A line annotation.
    """

    shape = 'line'
    """
    The name of the shape (:class:`str`).
    """

    color = ColorWithOpacity('white',
                             help='Assigns color for the line '
                                  '(`str` or `tuple`)').tag(wwt='lineColor')
    width = AstropyQuantity(1*u.pixel,
                            help='Assigns width for the line in pixels '
                                 '(:class:`~astropy.units.Quantity`)').tag(wwt='lineWidth')

    @validate('width')
    def _validate_width(self, proposal):
        if proposal['value'].unit.is_equivalent(u.pixel):
            return proposal['value'].to(u.pixel)
        else:
            raise TraitError('width must be in pixel equivalent unit')

    def add_point(self, coord):
        """
        Add one or more points to a line object.

        Parameters
        ----------
        coord : `~astropy.units.Quantity`
            The coordinates of the desired point(s).
        """
        coord_icrs = coord.icrs
        if coord_icrs.isscalar: # if coord only has one point
            self.parent._send_msg(event='line_add_point', id=self.id,
                                  ra=coord_icrs.ra.degree,
                                  dec=coord_icrs.dec.degree)
        else:
            for point in coord_icrs:
                self.parent._send_msg(event='line_add_point', id=self.id,
                                      ra=point.ra.degree,
                                      dec=point.dec.degree)

    def remove(self):
        """
        Removes the specified annotation from the current view.
        """
        self.parent._send_msg(event='remove_annotation', id=self.id)

    def _on_trait_change(self, changed):
        if isinstance(changed['new'], u.Quantity):
            changed['new'] = changed['new'].value

        super(Line, self)._on_trait_change(changed)


class FieldOfView():
    """
    A collection of polygon annotations. Takes the name of a pre-loaded
    telescope and displays its field of view.
    """
    # a more efficient method than CircleCollection of changing trait values?

    def __init__(self, parent, telescope, center, rot, **kwargs):
        # make sure rot is astropy quantity in proper units for self._rotate()
        try:
            if not rot.unit.is_equivalent(u.deg):
                raise AttributeError
        except AttributeError:
            raise ValueError('rotate argument must be Astropy quantity with '
                             'degree or radian-equivalent units')

        # get the JSON list of FOVs from parent BaseWWTWidget class
        self.parent = parent
        self._available = self.parent.instruments.available
        self._entry = self.parent.instruments.entry

        # list of IDs of annotations created in this FieldofView instance
        self.active = []
        self._gen_fov(telescope, center, rot, **kwargs)

    def _gen_fov(self, telescope, center, rot, **kwargs):
        # test if telescope is available
        if telescope not in self._available:
            raise ValueError('the given telescope\'s field of view is '
                             'unavailable at this time')

        position = self._entry[telescope][0]
        dimensions = self._entry[telescope][-1]
        # dimensions is a list of lists, i.e. [ [[ras], [decs]], '', ...]

        # test that pos matches what the user entered
        if center and position == 'absolute':
            raise ValueError('the given telescope does not take center '
                             'coordinates')
        elif not center and position == 'relative':
            raise ValueError('the given telescope requires center coordinates')

        for panel in dimensions:
            ras = (panel[0] * u.deg).value
            decs = (panel[1] * u.deg).value

            # rotate points if necessary
            if position == 'relative' and rot != 0:
                ras, decs = self._rotate(ras, decs, rot)

            # translate points to user-specified location if relative
            if position == 'relative':
                center_ra = center.ra.to(u.deg).value
                center_dec = center.dec.to(u.deg).value

                decs = center_dec + decs
                # scale RA by dec (due to polar contraction of spherical coords)
                ras = center_ra + ras / np.cos(decs * u.deg).value

                # check that abs(dec) < 90. if not, adjust it, and then
                # set the corresponding ra to be swapped by 180
                for i, dec in enumerate(decs):
                    if abs(dec) > 90:
                        if dec < -90:
                            decs[i] = -90. - (dec + 90.)
                        else: # dec > 90
                            decs[i] = 90. - (dec - 90.)
                        ras[i] += 180.

            corners = SkyCoord(ras, decs, unit=u.deg)
            if self.parent.galactic_mode:
                corners = corners.galactic

            # draw the panel
            annot = self.parent.add_polygon(corners, **kwargs)
            self.active.append(annot)

    def _rotate(self, ras, decs, rot):
        cos, sin = np.cos(rot).value, np.sin(rot).value
        ra_rot = ras * cos - decs * sin
        dec_rot = ras * sin + decs * cos

        return ra_rot, dec_rot

    def remove(self):
        """
        Removes the specified field of view from the viewer.
        """
        for annot in self.active:
            annot.remove()


class CircleCollection():
    """
    A collection of circle annotations. Takes a set of several points
    (e.g. a column of SkyCoords from an astropy Table) to make generating
    several circles at once easier.
    """

    def __init__(self, parent, points, **kwargs):
        if len(points) <= 1e4:
            self.points = points
        else:
            raise IndexError('For performance reasons, only 10,000 '
                        'annotations can be added at once for the time being.')
        self.parent = parent
        self.collection = []
        self._gen_circles(self.points, **kwargs)

    def _set_all_attributes(self, name, value):
        for elem in self.collection:
            setattr(elem, name, value)

    def _get_all_attributes(self, name):
        values = []
        for elem in self.collection:
            attr = getattr(elem, name)
            if attr not in values:
                values.append(attr)
        if len(values) == 1:
            return values[0]
        else:
            return values

    def _gen_circles(self, points, **kwargs):
        for elem in points:
            circle = Circle(self.parent, **kwargs)
            circle.set_center(elem)
            self.collection.append(circle)

    def add_points(self, points, **kwargs):
        """
        Adds multiple points to the CircleCollection.
        """
        self._gen_circles(points, **kwargs)
        self.points = concatenate((self.points, points))

    def remove(self):
        """
        Removes all circles in the CircleCollection from view.
        """
        for elem in self.collection:
            elem.remove()

    # Circle.__dict__ attributes
    @property
    def fill(self):
        """
        Whether or not the circles in the CircleCollection have a fill.
        """
        return self._get_all_attributes('fill')

    @fill.setter
    def fill(self, value):
        return self._set_all_attributes('fill', value)

    @property
    def fill_color(self):
        """
        The fill color of the circles in the CircleCollection.
        """
        return self._get_all_attributes('fill_color')

    @fill_color.setter
    def fill_color(self, value):
        return self._set_all_attributes('fill_color', value)

    @property
    def line_color(self):
        """
        The line color of the circles in the CircleCollection.
        """
        return self._get_all_attributes('line_color')

    @line_color.setter
    def line_color(self, value):
        return self._set_all_attributes('line_color', value)

    @property
    def line_width(self):
        """
        The line width of the circles in the CircleCollection.
        """
        return self._get_all_attributes('line_width')

    @line_width.setter
    def line_width(self, value):
        return self._set_all_attributes('line_width', value)

    @property
    def radius(self):
        """
        The radii of the circles in the CircleCollection.
        """
        return self._get_all_attributes('radius')

    @radius.setter
    def radius(self, value):
        return self._set_all_attributes('radius', value)

    @property
    def shape(self):
        """
        The shapes comprising the CircleCollection (always 'circle').
        """
        return self._get_all_attributes('shape')

    # Annotation.__dict__ attributes
    @property
    def label(self):
        """
        Descriptive text for the CircleCollection.
        """
        return self._get_all_attributes('label')

    @label.setter
    def label(self, value):
        return self._set_all_attributes('label', value)

    @property
    def hover_label(self):
        """
        Whether to show a label when the mouse hovers over the CircleCollection.
        """
        return self._get_all_attributes('hover_label')

    @hover_label.setter
    def hover_label(self, value):
        return self._set_all_attributes('hover_label', value)

    @property
    def opacity(self):
        """
        The opacity of the circles in the CircleCollection.
        """
        return self._get_all_attributes('opacity')

    @opacity.setter
    def opacity(self, value):
        return self._set_all_attributes('opacity', value)

    @property
    def tag(self):
        """
        A string that can be used by the web client.
        """
        return self._get_all_attributes('tag')

    @tag.setter
    def tag(self, value):
        return self._set_all_attributes('tag', value)



import os
from datetime import datetime

import pytest

try:
    import qtpy
except ImportError:
    QT_INSTALLED = False
else:
    QT_INSTALLED = True

try:
    import OpenGL.GL as gl
except ImportError:
    OPENGL_INSTALLED = False
else:
    OPENGL_INSTALLED = True

if QT_INSTALLED and OPENGL_INSTALLED:

    from qtpy.QtWidgets import QOpenGLWidget

    class OpenGLWidget(QOpenGLWidget):

        def getOpenGLInfo(self):

            gl_parameters = {'vendor': gl.GL_VENDOR,
                             'renderer': gl.GL_RENDERER,
                             'version': gl.GL_VERSION,
                             'shader': gl.GL_SHADING_LANGUAGE_VERSION}

            info = {}
            for key, setting in gl_parameters.items():
                try:
                    info[key] = gl.glGetString(setting).decode('ascii')
                except Exception:
                    info[key] = 'could not be determined'

            return info


def pytest_report_header(config):

    lines = []

    if QT_INSTALLED:

        # Get Python Qt bindings version
        lines.append("PyQt: {0}".format(qtpy.PYQT_VERSION))
        lines.append("PySide: {0}".format(qtpy.PYSIDE_VERSION))

        # Get WebEngine/WebKit info
        from qtpy.QtWebEngineWidgets import WEBENGINE
        lines.append("Web Framework: {0}".format('WebEngine' if WEBENGINE else 'WebKit'))

        if OPENGL_INSTALLED:

            # Get OpenGL version
            from .app import get_qapp
            get_qapp()
            widget = OpenGLWidget()
            widget.show()
            opengl_info = widget.getOpenGLInfo()
            widget.close()

            lines.append("OpenGL Vendor: {0}".format(opengl_info['vendor']))
            lines.append("OpenGL Renderer: {0}".format(opengl_info['renderer']))
            lines.append("OpenGL Version: {0}".format(opengl_info['version']))
            lines.append("Shader Version: {0}".format(opengl_info['shader']))

        else:

            lines.append("Could not determine OpenGL version (OpenGL package required)")

    else:

        lines.append("Qt not installed")

    return os.linesep + os.linesep.join(lines) + os.linesep


def pytest_unconfigure(config):

    if QT_INSTALLED:
        from .app import cleanup_qapp
        cleanup_qapp()


REFERENCE_TIME = datetime(2017, 1, 1, 0, 0, 0, 0)


@pytest.fixture(scope='session')
def wwt_qt_client():
    from .qt import WWTQtClient
    wwt = WWTQtClient(block_until_ready=True, size=(400, 400))
    return wwt


@pytest.fixture(scope='function')
def wwt_qt_client_isolated():
    from .qt import WWTQtClient
    wwt = WWTQtClient(block_until_ready=True, size=(400, 400))
    return wwt


@pytest.fixture(autouse=True)
def reset_state(wwt_qt_client):
    wwt_qt_client.reset()
    wwt_qt_client.pause_time()
    wwt_qt_client.set_current_time(REFERENCE_TIME)
    yield  # This yields the test itself



import os
from traitlets import HasTraits, observe, validate, TraitError
from astropy import units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord

# We import the trait classes from .traits since we do various customizations
from .traits import Color, Bool, Float, Unicode, AstropyQuantity

from .annotation import Circle, Polygon, Line, FieldOfView, CircleCollection
from .imagery import get_imagery_layers, ImageryLayers
from .solar_system import SolarSystem
from .layers import LayerManager
from .instruments import Instruments

# The WWT web control API is described here:
# https://worldwidetelescope.gitbook.io/html5-control-reference/

DEFAULT_SURVEYS_URL = 'https://WorldWideTelescope.github.io/pywwt/surveys.xml'

VIEW_MODES_2D = ['sky', 'sun', 'mercury', 'venus', 'earth', 'moon', 'mars',
                 'jupiter', 'callisto', 'europa', 'ganymede', 'io',
                 'saturn', 'uranus', 'neptune', 'pluto',
                 'panorama']
VIEW_MODES_3D = ['solar system', 'milky way', 'universe']


__all__ = ['BaseWWTWidget']


class BaseWWTWidget(HasTraits):
    """
    The core class in common to the Qt and Jupyter widgets.

    This class provides a common interface to modify settings and interact with
    WorldWide Telescope.
    """
    def __init__(self, **kwargs):
        super(BaseWWTWidget, self).__init__()
        self.observe(self._on_trait_change, type='change')
        self._available_layers = get_imagery_layers(DEFAULT_SURVEYS_URL)
        self.imagery = ImageryLayers(self._available_layers)
        self.solar_system = SolarSystem(self)
        self._instruments = Instruments()
        self.current_mode = 'sky'
        self._paused = False
        self.layers = LayerManager(parent=self)

        # NOTE: we deliberately don't force _on_trait_change to be called here
        # for the WWT settings, as the default values are hard-coded in wwt.html
        # This is done because there is otherwise no reliable way of making sure
        # that we would call _on_trait_change once WWT is ready to receive
        # commands. There is a test in test_core.py that ensures that the
        # defaults here are in sync with the defaults in wwt.html

    def _on_trait_change(self, changed):
        # This method gets called anytime a trait gets changed. Since this class
        # gets inherited by the Jupyter widgets class which adds some traits of
        # its own, we only want to react to changes in traits that have the wwt
        # metadata attribute (which indicates the name of the corresponding WWT
        # setting).
        wwt_name = self.trait_metadata(changed['name'], 'wwt')
        new_value = changed['new']
        if wwt_name is not None:
            if isinstance(new_value, u.Quantity):
                new_value = new_value.value

            self._send_msg(event='setting_set',
                           setting=wwt_name,
                           value=new_value)

    def _send_msg(self, **kwargs):
        # This method should be overridden and should send the message to WWT
        pass
    
    def _get_view_data(self, field):
        # This method should be overwritten to get the RA, Dec, and FoV of the current view
        pass

    actual_planet_scale = Bool(False,
                               help='Whether to show planets to scale or as '
                                    'points with a fixed size '
                                    '(`bool`)').tag(wwt='actualPlanetScale')

    # TODO: need to add all settings as traits
    # check wwt.html for comments on settings that are disabled below

    constellation_boundary_color = Color('blue',
                                         help='The color of the constellation '
                                         'boundaries (`str` or '
                                         '`tuple`)').tag(wwt='constellationBoundryColor')
    constellation_figure_color = Color('red',
                                       help='The color of the constellation '
                                            'figure (`str` or '
                                            '`tuple`)').tag(wwt='constellationFigureColor')
    constellation_selection_color = Color('yellow',
                                          help='The color of the constellation '
                                               'selection (`str` or '
                                               '`tuple`)').tag(wwt='constellationSelectionColor')

    constellation_boundaries = Bool(False,
                                    help='Whether to show boundaries for the '
                                         'selected constellations '
                                         '(`bool`)').tag(wwt='showConstellationBoundries')
    constellation_figures = Bool(False,
                                 help='Whether to show the constellations '
                                      '(`bool`)').tag(wwt='showConstellationFigures')
    constellation_selection = Bool(False,
                                   help='Whether to only show boundaries for '
                                        'the selected constellation '
                                        '(`bool`)').tag(wwt='showConstellationSelection')
    #constellation_pictures = Bool(False, help='Whether to show pictures of the constellations\' mythological representations (`bool`)').tag(wwt='showConstellationPictures')
    #constellation_labels = Bool(False, help='Whether to show labelss for constellations (`bool`)').tag(wwt='showConstellationLabels')

    crosshairs = Bool(False, help='Whether to show crosshairs at the center of '
                                 'the field (`bool`)').tag(wwt='showCrosshairs')
    crosshairs_color = Color('white',
                             help='The color of the crosshairs '
                                  '(`str` or `tuple`)').tag(wwt='crosshairsColor')
    grid = Bool(False, help='Whether to show the equatorial grid '
                            '(`bool`)').tag(wwt='showGrid')
    ecliptic = Bool(False, help='Whether to show the path of the ecliptic '
                                '(`bool`)').tag(wwt='showEcliptic')
    ecliptic_grid = Bool(False, help='Whether to show a grid relative to the '
                                     'ecliptic plane (`bool`)').tag(wwt='showEclipticGrid')

    # TODO: need to add more methods here.

    def clear_annotations(self):
        """
        Clears all annotations from the current view.
        """
        return self._send_msg(event='clear_annotations')

    def get_center(self):
        """
        Return the view's current right ascension and declination in degrees.
        """
        return SkyCoord(self._get_view_data('ra'),
                        self._get_view_data('dec'),
                        unit=(u.hourangle, u.deg))

    def get_fov(self):
        """
        Return the view's current field of view in degrees
        """
        return self._get_view_data('fov') * u.deg

    def load_tour(self, url):
        """
        Load and begin playing a tour based on the URL to a .wtt file from
        the WorldWideTelescope website.

        Parameters
        ----------
        url : `str`
            The URL of the chosen tour -- must be a .wtt file.
        """
        # throw error if url doesn't end in .wtt
        if url[-4:] == '.wtt':
            self._send_msg(event='load_tour', url=url)
        else:
            raise ValueError('url must end in \'.wwt\'')

    def pause_tour(self):
        """
        Pause a loaded tour.
        """
        self._send_msg(event='pause_tour')

    def resume_tour(self):
        """
        Resume a paused tour.
        """
        self._send_msg(event='resume_tour')

    def pause_time(self):
        """
        Pause the progression of time in the viewer.
        """
        self._send_msg(event='pause_time')

    def play_time(self, rate=1):
        """
        Resume the progression of time in the viewer.

        Parameters
        ----------
        rate : int or float
            The rate at which time passes (1 meaning real-time)
        """
        self._send_msg(event='resume_time', rate=rate)

    def set_current_time(self, dt=None):
        """
        Set the viewer time to match the real-world time.

        Parameters
        ----------
        dt : `~datetime.datetime` or `~astropy.time.Time`
            The current time, either as a `datetime.datetime` object or an
            astropy :class:`astropy.time.Time` object. If not specified, this
            uses the current time
        """
        if dt is None:
            dt = Time.now()
        if isinstance(dt, Time):
            dt = dt.datetime
        self._send_msg(event='set_datetime',
                       year=dt.year, month=dt.month, day=dt.day,
                       hour=dt.hour, minute=dt.minute, second=dt.second,
                       millisecond=int(dt.microsecond / 1000.))

    def center_on_coordinates(self, coord, fov=60 * u.deg, instant=True):
        """
        Center the view on a particular object or point in the sky.

        Parameters
        ----------
        coord : `~astropy.units.Quantity`
            The set of coordinates the view should center on.

        fov : `~astropy.units.Quantity`, optional
            The desired field of view.

        instant : `bool`, optional
            Whether the view changes instantly or smoothly scrolls to the
            desired location.
        """
        coord_icrs = coord.icrs
        self._send_msg(event='center_on_coordinates',
                       ra=coord_icrs.ra.deg,
                       dec=coord_icrs.dec.deg,
                       fov=fov.to(u.deg).value,
                       instant=instant)

    galactic_mode = Bool(False,
                         help='Whether the galactic plane should be horizontal '
                              'in the viewer (`bool`)').tag(wwt='galacticMode')
    galactic_grid = Bool(False, help='Whether to show a grid relative to the '
                                     'galactic plane (`bool`)').tag(wwt='showGalacticGrid')
    #galactic_text = Bool(False, help='Whether to show labels for the galactic grid\'s text (`bool`)').tag(wwt='showGalacticGridText')
    alt_az_grid = Bool(False, help='Whether to show an altitude-azimuth grid '
                                   '(`bool`)').tag(wwt='showAltAzGrid')
    #alt_az_text = Bool(False, help='Whether to show labels for the altitude-azimuth grid\'s text (`bool`)').tag(wwt='showAltAzGridText')

    local_horizon_mode = Bool(False, help='Whether the view should be that of '
                                          'a local latitude, longitude, and '
                                          'altitude (`bool`)').tag(wwt='localHorizonMode')
    location_altitude = AstropyQuantity(0 * u.m,
                                        help='The altitude of the viewing '
                                             'location in local horizon mode '
                                             '(:class:`~astropy.units.Quantity`)').tag(wwt='locationAltitude')
    location_latitude = AstropyQuantity(47.633 * u.deg,
                                        help='The latitude of the viewing '
                                              'location in local horizon mode '
                                              '(:class:`~astropy.units.Quantity`)').tag(wwt='locationLat')
    location_longitude = AstropyQuantity(122.133333 * u.deg,
                                         help='The longitude of the viewing '
                                              'location in local horizon mode '
                                              '(:class:`~astropy.units.Quantity`)').tag(wwt='locationLng')

    @validate('location_altitude')
    def _validate_altitude(self, proposal):
        if proposal['value'].unit.physical_type == 'length':
            return proposal['value'].to(u.meter)
        else:
            raise TraitError('location_altitude not in units of length')

    @validate('location_latitude')
    def _validate_latitude(self, proposal):
        if proposal['value'].unit.physical_type == 'angle':
            return proposal['value'].to(u.degree)
        else:
            raise TraitError('location_latitude not in angle units')

    @validate('location_longitude')
    def _validate_longitude(self, proposal):
        if proposal['value'].unit.physical_type == 'angle':
            return proposal['value'].to(u.degree)
        else:
            raise TraitError('location_longitude not in angle units')

    def set_view(self, mode):
        """
        Change the view mode.

        Valid options include the default sky mode, a 3D universe mode with
        different viewing levels (the solar system, the Milky Way, and the
        observed universe), individual views of major solar system objects, and
        panoramas from lunar missions and NASA's Mars rovers.

        To find the list of available views, use the
        :attr:`~pywwt.BaseWWTWidget.available_views`.

        Parameters
        ----------
        mode : `str`
            The desired view mode. (default: 'sky')
        """

        mode = mode.lower()

        solar_system_mode = '3D Solar System View'

        if mode in VIEW_MODES_2D:
            if mode == 'earth':
                # Switch to a daytime view of the earth
                mode = 'Bing Maps Aerial'
            elif mode == 'mars':
                mode = 'Visible Imagery'
            self._send_msg(event='set_viewer_mode', mode=mode)
            if mode == 'sky' or mode == 'panorama':
                self.current_mode = mode
            else:
                self.current_mode = 'planet'
        elif mode in VIEW_MODES_3D:
            self._send_msg(event='set_viewer_mode', mode=solar_system_mode)
            self.current_mode = mode
        else:
            raise ValueError('mode should be one of {0}'.format('/'.join(VIEW_MODES_2D + VIEW_MODES_3D)))

        self.reset_view()

    def reset_view(self):
        """
        Reset the current view mode's coordinates and field of view to
        their original states.
        """
        if self.current_mode == 'sky':
            self.center_on_coordinates(SkyCoord(0., 0., unit=u.deg),
                                       fov=60*u.deg, instant=False)
        if self.current_mode == 'planet':
            self.center_on_coordinates(SkyCoord(35.55, 11.43, unit=u.deg),
                                       fov=40*u.deg, instant=False)
        if self.current_mode == 'solar system':
            self.center_on_coordinates(SkyCoord(0., 0., unit=u.deg),
                                       fov=50*u.deg, instant=False)
        if self.current_mode == 'milky way':
            self.center_on_coordinates(SkyCoord(114.85, -29.52, unit=u.deg),
                                       fov=6e9*u.deg, instant=False)
        if self.current_mode == 'universe':
            self.center_on_coordinates(SkyCoord(16.67, 37.72, unit=u.deg),
                                       fov=1e14*u.deg, instant=False)
        if self.current_mode == 'panorama':
            pass

    @property
    def available_views(self):
        """
        A list of the modes that are currently available in the viewer.
        """
        return sorted(VIEW_MODES_2D + VIEW_MODES_3D)

    def load_image_collection(self, url):
        """
        Load a collection of layers for possible use in the viewer.

        Parameters
        ----------
        url : `str`
            The URL of the desired image collection.
        """
        self._available_layers.update(get_imagery_layers(url))
        self._send_msg(event='load_image_collection', url=url)

    @property
    def available_layers(self):
        """
        A list of the layers that are currently available in the viewer.
        """
        return sorted(self._available_layers)

    foreground = Unicode('Digitized Sky Survey (Color)',
                         help='The layer to show in the foreground (`str`)')

    @observe('foreground')
    def _on_foreground_change(self, changed):
        self._send_msg(event='set_foreground_by_name', name=changed['new'])
        # Changing a layer resets the opacity, so we re-trigger the opacity setting
        self._send_msg(event='set_foreground_opacity',
                       value=self.foreground_opacity * 100)

    @validate('foreground')
    def _validate_foreground(self, proposal):
        if proposal['value'] in self.available_layers:
            return proposal['value']
        else:
            raise TraitError('foreground is not one of the available layers')

    background = Unicode('Hydrogen Alpha Full Sky Map',
                         help='The layer to show in the background (`str`)')

    @observe('background')
    def _on_background_change(self, changed):
        self._send_msg(event='set_background_by_name', name=changed['new'])
        # Changing a layer resets the opacity, so we re-trigger the opacity setting
        self._send_msg(event='set_foreground_opacity',
                       value=self.foreground_opacity * 100)

    @validate('background')
    def _validate_background(self, proposal):
        if proposal['value'] in self.available_layers:
            return proposal['value']
        else:
            raise TraitError('background is not one of the available layers')

    foreground_opacity = Float(0.8, help='The opacity of the foreground layer '
                                         '(`float`)')

    @observe('foreground_opacity')
    def _on_foreground_opacity_change(self, changed):
        self._send_msg(event='set_foreground_opacity',
                       value=changed['new'] * 100)

    @validate('foreground_opacity')
    def _validate_foreground_opacity(self, proposal):
        if 0 <= proposal['value'] <= 1:
            return proposal['value']
        else:
            raise TraitError('foreground_opacity should be between 0 and 1')

    def add_circle(self, center=None, **kwargs):
        """
        Add a circle annotation to the current view.

        Parameters
        ----------
        center : `~astropy.units.Quantity`, optional
            The coordinates of desired center of the circle. If blank,
            defaults to the center of the current view.
        kwargs
            Optional arguments that allow corresponding Circle or Annotation
            attributes to be set upon shape initialization.
        """
        # TODO: could buffer JS call here
        circle = Circle(parent=self, **kwargs)
        if center:
            circle.set_center(center)
        return circle

    def add_polygon(self, points=None, **kwargs):
        """
        Add a polygon annotation to the current view.

        Parameters
        ----------
        points : `~astropy.units.Quantity`, optional
            The desired points that make up the polygon. If blank or just
            one point, the annotation will be initialized but will not be
            visible until more points are added. Note that the points should
            be specified in counter-clockwise order on the sky if you intend
            to fill the polygon.
        kwargs
            Optional arguments that allow corresponding Polygon or
            Annotation attributes to be set upon shape initialization.
        """
        # same TODO as above
        polygon = Polygon(parent=self, **kwargs)
        if points:
            polygon.add_point(points)
        return polygon

    def add_line(self, points=None, **kwargs):
        """
        Add a line annotation to the current view.

        Parameters
        ----------
        points : `~astropy.units.Quantity`, optional
            The desired points that make up the line. If blank or just one
            point, the annotation will be initialized but will not be
            visible until more points are added.
        kwargs
            Optional arguments that allow corresponding Line or Annotation
            attributes to be set upon shape initialization.
        """
        # same TODO as above
        line = Line(parent=self, **kwargs)
        if points:
            line.add_point(points)
        return line

    @property
    def instruments(self):
        """
        Instruments available for use in `add_fov`
        """
        return self._instruments

    def add_fov(self, telescope, center=None, rotate=0*u.rad, **kwargs):
        """
        Add a telescope's field of view (FOV) to the current view.

        Parameters
        ----------
        telescope : `str`
            The telescope whose field of view will be displayed. Be sure to use
            the ``instruments`` attribute to see and select from the preset list
            of instruments available in pyWWT.
        center : `~astropy.units.Quantity`, optional
            The coordinates of desired center of the FOV. If blank,
            defaults to the center of the current view.
        rotate : `~astropy.units.Quantity`, optional
            The amount to rotate the FOV. Both radians and degrees are
            accepted. If blank, defaults to 0 radians (no rotation).
        kwargs
            Optional arguments that allow corresponding Polygon or
            Annotation attributes to be set upon shape initialization.
        """
        return FieldOfView(self, telescope, center, rotate, **kwargs)

    def add_collection(self, points, **kwargs):
        """
        Add a CircleCollection to the current view.

        Parameters
        ----------
        points : `~astropy.units.Quantity`
            The desired points that will serve as the centers of the
            circles that make up the collection. Requires at least two
            sets of coordinates for initialization.
        kwargs
            Optional arguments that allow corresponding Circle or
            Annotation attributes to be set upon shape initialization.
        """
        collection = CircleCollection(self, points, **kwargs)
        return collection

    def reset(self):
        """
        Reset WWT to initial state.
        """

        # Remove any existing layers
        for layer in self.layers:
            layer.remove()

        # Reset coordinates to initial view
        gc = SkyCoord(0, 0, unit=('deg', 'deg'), frame='icrs')
        self.center_on_coordinates(gc, 60 * u.deg)

        # Reset traits to default values
        for trait_name, trait in self.traits().items():
            setattr(self, trait_name, trait.default_value)



from traitlets import (TraitType, TraitError,
                       Any as OriginalAny,
                       Bool as OriginalBool,
                       CFloat as OriginalFloat,
                       Int as OriginalInt,
                       Unicode as OriginalUnicode)
from astropy import units as u
import six

try:
    from matplotlib.colors import to_hex
except ImportError:
    from matplotlib.colors import colorConverter, rgb2hex
    def to_hex(input):  # noqa
        return rgb2hex(colorConverter.to_rgb(input))

__all__ = ['Any', 'Bool', 'Float', 'Int', 'Unicode', 'AstropyQuantity',
           'Color', 'ColorWithOpacity', 'to_hex']

# We inherit the original trait classes to make sure that the docstrings are set


class Any(OriginalAny):

    def __init__(self, *args, **kwargs):
        super(Any, self).__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help


class Bool(OriginalBool):

    def __init__(self, *args, **kwargs):
        super(Bool, self).__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help


class Float(OriginalFloat):

    def __init__(self, *args, **kwargs):
        super(Float, self).__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help


class Int(OriginalInt):

    def __init__(self, *args, **kwargs):
        super(Int, self).__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help


class Unicode(OriginalUnicode):

    def __init__(self, *args, **kwargs):
        super(Unicode, self).__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help


class AstropyQuantity(TraitType):

    default = 0 * u.one
    info_text = '\'Custom trait to handle astropy quantities with units\''

    def __init__(self, *args, **kwargs):
        super(AstropyQuantity, self).__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help

    def validate(self, obj, value):
        if isinstance(value, u.Quantity):
            return value
        self.error(obj, value)


class Color(TraitType):

    def __init__(self, *args, **kwargs):
        super(Color, self).__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help

    def validate(self, obj, value):
        if isinstance(value, six.string_types) or (isinstance(value, tuple) and len(value) == 3):
            return to_hex(value)
        else:
            if hasattr(obj, 'opacity'):
                raise TraitError('color must be a string or a tuple of 3 or 4 floats')
            else:
                raise TraitError('color must be a string or a tuple of 3 floats')


class ColorWithOpacity(Color):

    def validate(self, obj, value):
        if isinstance(value, tuple) and len(value) == 4:
            obj.opacity = value[-1]
            value = value[:3]
        return super(ColorWithOpacity, self).validate(obj, value)



import os
import pytest

import numpy as np

from astropy.wcs import WCS
from astropy.table import Table
from astropy import units as u
from astropy.coordinates import SkyCoord

from .test_qt_widget import assert_widget_image

from ..core import BaseWWTWidget
from ..layers import TableLayer, guess_lon_lat_columns, csv_table_win_newline


class TestLayers:

    def setup_method(self, method):
        self.client = BaseWWTWidget()
        self.table = Table()
        self.table['flux'] = [2, 3, 4]
        self.table['dec'] = [4, 5, 6]
        self.table['ra'] = [1, 2, 3] * u.deg

    def test_add_and_remove_layer(self, capsys):

        assert len(self.client.layers) == 0
        assert str(self.client.layers) == 'Layer manager with no layers'
        layer1 = self.client.layers.add_table_layer(table=self.table)
        assert len(self.client.layers) == 1
        assert str(self.client.layers) == ('Layer manager with 1 layers:\n\n'
                                           '  [0]: TableLayer with 3 markers\n')

        layer2 = self.client.layers.add_table_layer(table=self.table)

        assert len(self.client.layers) == 2
        assert str(self.client.layers) == ('Layer manager with 2 layers:\n\n'
                                           '  [0]: TableLayer with 3 markers\n'
                                           '  [1]: TableLayer with 3 markers\n')

        assert self.client.layers[0] is layer1
        assert self.client.layers[1] is layer2

        # Test iteration
        for layer in self.client.layers:
            assert isinstance(layer, TableLayer)

        layer1.remove()

        assert len(self.client.layers) == 1
        assert str(self.client.layers) == ('Layer manager with 1 layers:\n\n'
                                           '  [0]: TableLayer with 3 markers\n')

        self.client.layers.remove_layer(layer2)

        assert len(self.client.layers) == 0
        assert str(self.client.layers) == 'Layer manager with no layers'

    def test_alt_unit(self):

        layer = self.client.layers.add_table_layer(table=self.table)

        # Using a string
        layer.alt_unit = 'm'

        # Using a string of an imperial unit
        layer.alt_unit = 'inch'

        # Using an astropy unit
        layer.alt_unit = u.km

        # Using a unit that is equal but not identical to one of the accepted ones
        layer.alt_unit = u.def_unit('same_as_km', 1000 * u.m)

        # Using an invalid string
        with pytest.raises(ValueError) as exc:
            layer.alt_unit = 'banana'
        assert exc.value.args[0].strip() == "'banana' did not parse as unit: At col 0, banana is not a valid unit."

        # Using an unsupported unit
        with pytest.raises(ValueError) as exc:
            layer.alt_unit = u.kg
        assert exc.value.args[0].strip() == "alt_unit should be one of AU/Mpc/ft/inch/km/lyr/m/mi/pc"

        # Using a non-equal custom unit
        with pytest.raises(ValueError) as exc:
            layer.alt_unit = u.def_unit('same_as_half_km', 500 * u.m)
        assert exc.value.args[0].strip() == "alt_unit should be one of AU/Mpc/ft/inch/km/lyr/m/mi/pc"

    def test_lon_unit(self):

        layer = self.client.layers.add_table_layer(table=self.table)

        # Using a string
        layer.lon_unit = 'deg'

        # Using an astropy unit
        layer.lon_unit = u.hourangle

        # Using a unit that is equal but not identical to one of the accepted ones
        layer.lon_unit = u.def_unit('same_as_deg', 3600 * u.arcsec)

        # Using an invalid string
        with pytest.raises(ValueError) as exc:
            layer.lon_unit = 'banana'
        assert exc.value.args[0].strip() == "'banana' did not parse as unit: At col 0, banana is not a valid unit."

        # Using an unsupported unit
        with pytest.raises(ValueError) as exc:
            layer.lon_unit = u.kg
        assert exc.value.args[0].strip() == "lon_unit should be one of deg/h/hourangle"

        # Using a non-equal custom unit
        with pytest.raises(ValueError) as exc:
            layer.lon_unit = u.def_unit('same_as_arcmin', 60 * u.arcsec)
        assert exc.value.args[0].strip() == "lon_unit should be one of deg/h/hourangle"

    def test_alt_type(self):

        layer = self.client.layers.add_table_layer(table=self.table)

        layer.alt_type = 'depth'

        with pytest.raises(ValueError) as exc:
            layer.alt_type = 'time'
        assert exc.value.args[0].strip() == "alt_type should be one of depth/altitude/distance/seaLevel/terrain"

    def test_auto_alt_unit(self):

        self.table['altitude'] = [1, 4, 5] * u.au
        self.table['altitude2'] = [1, 4, 5] * u.def_unit('same_as_km', 1000 * u.m)
        self.table['flux'].unit = u.kg

        layer = self.client.layers.add_table_layer(table=self.table)

        assert layer.alt_att == ''
        assert layer.alt_unit is None

        layer.alt_att = 'altitude'
        assert layer.alt_unit is u.au

        layer.alt_att = 'altitude2'
        assert layer.alt_unit is u.km

        expected_warning = ('Column flux has units of kg but this is not a '
                            'valid unit of altitude - set the unit directly with '
                            'alt_unit')

        with pytest.warns(UserWarning, match=expected_warning):
            layer.alt_att = 'flux'

    def test_auto_lon_unit(self):

        self.table['longitude'] = [1, 4, 5] * u.hour
        self.table['longitude2'] = [1, 4, 5] * u.def_unit('same_as_deg', 3600 * u.arcsec)
        self.table['flux'].unit = u.kg

        layer = self.client.layers.add_table_layer(table=self.table)

        assert layer.lon_att == 'ra'
        assert layer.lon_unit is u.deg

        layer.lon_att = 'longitude'
        assert layer.lon_unit is u.hour

        layer.lon_att = 'longitude2'
        assert layer.lon_unit is u.deg

        expected_warning = ('Column flux has units of kg but this is not a '
                            'valid unit of longitude - set the unit directly with '
                            'lon_unit')

        with pytest.warns(UserWarning, match=expected_warning):
            layer.lon_att = 'flux'

    def test_update_data(self):

        self.table['flux'].unit = 'm'
        layer = self.client.layers.add_table_layer(table=self.table,
                                                  lon_att='ra', lat_att='dec', alt_att='flux')

        assert layer.lon_att == 'ra'
        assert layer.lon_unit is u.deg
        assert layer.lat_att == 'dec'
        assert layer.alt_att == 'flux'
        assert layer.alt_unit is u.m

        # Replace with a table with the same column names but different units
        # for the lon and alt
        table = Table()
        table['ra'] = [1, 2, 3] * u.hourangle
        table['dec'] = [4, 5, 6]
        table['flux'] = [2, 3, 4] * u.km
        layer.update_data(table=table)

        assert layer.lon_att == 'ra'
        assert layer.lon_unit is u.hourangle
        assert layer.lat_att == 'dec'
        assert layer.alt_att == 'flux'
        assert layer.alt_unit is u.km

        # Replace with a table with different column names
        table = Table()
        table['a'] = [1, 2, 3] * u.deg
        table['b'] = [4, 5, 6]
        table['c'] = [2, 3, 4] * u.au
        layer.update_data(table=table)

        assert layer.lon_att == 'a'
        assert layer.lon_unit is u.deg
        assert layer.lat_att == 'b'
        assert layer.alt_att == ''

    def test_line_endings(self):
        self.table['ra'] = [1,2,3]
        expected_str = "flux,dec,ra\r\n"\
                        "2,4,1\r\n"\
                        "3,5,2\r\n"\
                        "4,6,3\r\n"
        assert csv_table_win_newline(self.table) == expected_str

    def test_deprecated_api_call(self, capsys):
        """For the time being, test that the deprecated name for this function still
        works, but issues a warning

        """
        import warnings

        assert len(self.client.layers) == 0
        assert str(self.client.layers) == 'Layer manager with no layers'

        with warnings.catch_warnings(record=True) as w:
            layer1 = self.client.layers.add_data_layer(table=self.table)

        assert len(self.client.layers) == 1
        assert str(self.client.layers) == ('Layer manager with 1 layers:\n\n'
                                           '  [0]: TableLayer with 3 markers\n')
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)


CASES = [[('flux', 'dec', 'ra'), ('ra', 'dec')],
         [('mass', 'lat', 'lon'), ('lon', 'lat')],
         [('a', 'lng', 'b', 'lat'), ('lng', 'lat')],
         [('flux', 'ra', 'radius', 'dec'), ('ra', 'dec')],
         [('FLUX', 'DECJ2000', 'RAJ2000'), ('RAJ2000', 'DECJ2000')],
         [('DISTANCE', 'LON1', 'LAT1'), ('LON1', 'LAT1')],
         [('flux', 'lng2', 'lat2', 'lng1', 'lat1'), (None, None)],
         [('ra', 'ra', 'dec'), (None, None)]]


@pytest.mark.parametrize(('colnames', 'expected'), CASES)
def test_guess_lon_lat_columns(colnames, expected):
    assert guess_lon_lat_columns(colnames) == expected


def test_layers_image(tmpdir, wwt_qt_client):

    # A series of tests that excercise the layer functionality and compare
    # the results with a set of baseline images.

    wwt = wwt_qt_client

    wwt.foreground = 'Black Sky Background'
    wwt.background = 'Black Sky Background'

    # TODO: need a way to completely turn off sun + planets. For now we just
    # point towards the ecliptic North pole
    wwt.center_on_coordinates(SkyCoord(18 * u.hourangle, 66 * u.deg))

    # Simple default case

    table = Table()
    table['flux'] = [2, 3, 4, 5, 6]
    table['dec'] = [84, 85, 86, 87, 88]
    table['ra'] = [250, 260, 270, 280, 290] * u.deg

    layer1 = wwt.layers.add_table_layer(table=table)

    # Case where we change the default values on initialization

    table = Table()
    table['flux'] = [2, 3, 4, 5, 6]
    table['dec'] = [74, 75, 76, 77, 78]
    table['ra'] = [250, 260, 270, 280, 290] * u.deg
    table['other'] = [255, 265, 275, 285, 295] * u.deg

    layer2 = wwt.layers.add_table_layer(table=table, color='red', lon_att='other', size_scale=100, opacity=0.5)

    # Case where we change the values after initialization

    table = Table()
    table['flux'] = [2, 3, 4, 5, 6]
    table['dec'] = [64, 65, 66, 67, 68]
    table['ra'] = [250, 260, 270, 280, 290] * u.deg
    table['other'] = [255, 265, 275, 285, 295] * u.deg

    layer3 = wwt.layers.add_table_layer(table=table)

    wwt.wait(2)

    layer3.color = 'green'
    layer3.lon_att = 'other'
    layer3.size_scale = 50
    layer3.opacity = 0.8

    # Case with size and color encoding where we change the default values on initialization

    table = Table()
    table['flux'] = [2, 3, 4, 5, 6]
    table['dec'] = [54, 55, 56, 57, 58]
    table['ra'] = [250, 260, 270, 280, 290] * u.deg
    table['other'] = [255, 265, 275, 285, 295] * u.deg

    layer4 = wwt.layers.add_table_layer(table=table, cmap_att='other', size_att='flux')

    # Case with size and color encoding where we change the values after initialization

    table = Table()
    table['flux'] = [2, 3, 4, 5, 6]
    table['dec'] = [44, 45, 46, 47, 48]
    table['ra'] = [250, 260, 270, 280, 290] * u.deg
    table['other'] = [255, 265, 275, 285, 295] * u.deg

    layer5 = wwt.layers.add_table_layer(table=table)

    wwt.wait(2)

    layer5.cmap_att = 'other'
    layer5.size_att = 'flux'

    wwt.wait(2)

    # For now this test doesn't work in CI, seemingly because of some
    # OpenGL features that aren't available there.
    if os.environ.get('CI', 'false').lower() == 'false':
        assert_widget_image(tmpdir, wwt, 'sky_layers.png')


def test_image_layer_equ(tmpdir, wwt_qt_client_isolated):

    # A series of tests that excercise the image layer functionality and compare
    # the results with a set of baseline images.

    # NOTE: due to an unknown issue, we need to run this using an isolated
    # Qt client and we can't re-use the usual wwt_qt_client fixture, as loading
    # the image layer appears to have some kind of irreversible impact on the
    # state of the Qt widget.

    wwt = wwt_qt_client_isolated

    wwt.foreground = 'Black Sky Background'
    wwt.background = 'Black Sky Background'

    wwt.center_on_coordinates(SkyCoord(30 * u.deg, 40 * u.deg))

    array = np.arange(10000).reshape((100, 100))
    wcs = WCS()
    wcs.wcs.ctype = 'RA---TAN', 'DEC--TAN'
    # wcs.wcs.ctype = 'GLON-CAR', 'GLAT-CAR'
    wcs.wcs.crval = 30, 40
    wcs.wcs.crpix = 50.5, 50.5
    wcs.wcs.cdelt = -0.1, 0.1

    wwt.layers.add_image_layer(image=(array, wcs))

    wwt.wait(2)

    # For now this test doesn't work in CI, seemingly because of some
    # OpenGL features that aren't available there.
    if os.environ.get('CI', 'false').lower() == 'false':
        assert_widget_image(tmpdir, wwt, 'image_layer_equ.png')


def test_image_layer_gal(tmpdir, wwt_qt_client_isolated):

    # A series of tests that excercise the image layer functionality and compare
    # the results with a set of baseline images.

    # NOTE: due to an unknown issue, we need to run this using an isolated
    # Qt client and we can't re-use the usual wwt_qt_client fixture, as loading
    # the image layer appears to have some kind of irreversible impact on the
    # state of the Qt widget.

    wwt = wwt_qt_client_isolated

    wwt.foreground = 'Black Sky Background'
    wwt.background = 'Black Sky Background'

    array = np.arange(10000).reshape((100, 100))
    wcs = WCS()
    wcs.wcs.ctype = 'GLON-CAR', 'GLAT-CAR'
    wcs.wcs.crpix = 50.5, 50.5
    wcs.wcs.cdelt = -0.03, 0.03

    wcs.wcs.crval = 33, 43
    wwt.layers.add_image_layer(image=(array, wcs))

    wcs.wcs.crval = 27, 43
    layer2 = wwt.layers.add_image_layer(image=(array, wcs))
    layer2.vmin = -5000
    layer2.vmax = 15000

    wcs.wcs.crval = 27, 37
    layer3 = wwt.layers.add_image_layer(image=(array, wcs))
    layer3.stretch = 'log'

    wcs.wcs.crval = 33, 37
    layer4 = wwt.layers.add_image_layer(image=(array, wcs))
    layer4.opacity = 0.5

    wwt.wait(2)

    wwt.center_on_coordinates(SkyCoord(30 * u.deg, 40 * u.deg, frame='galactic'), fov=14 * u.deg)

    wwt.wait(2)

    # For now this test doesn't work in CI, seemingly because of some
    # OpenGL features that aren't available there.
    if os.environ.get('CI', 'false').lower() == 'false':
        assert_widget_image(tmpdir, wwt, 'image_layer_gal.png')



import os
import json
import mimetypes
from hashlib import md5
from tornado import web
from notebook.utils import url_path_join
from notebook.base.handlers import IPythonHandler

__all__ = ['load_jupyter_server_extension']


STATIC_DIR = os.path.join(os.path.dirname(__file__), 'nbextension', 'static')
CONFIG = os.path.expanduser('~/.pywwt')


class WWTFileHandler(IPythonHandler):

    def get(self, filename):

        filename = os.path.basename(filename)

        # First we check if this is a standard file in the static directory
        if os.path.exists(os.path.join(STATIC_DIR, filename)):
            path = os.path.join(STATIC_DIR, filename)
        else:
            # If not, we open the config file which should contain a JSON
            # dictionary with filenames and paths.
            if not os.path.exists(CONFIG):
                raise web.HTTPError(404)
            with open(CONFIG) as f:
                config = json.load(f)
            if filename in config['paths']:
                path = config['paths'][filename]
            else:
                raise web.HTTPError(404)

        # Do our best to set an appropriate Content-Type.
        self.set_header('Content-Type', mimetypes.guess_type(filename)[0])

        with open(path, 'rb') as f:
            content = f.read()

        self.finish(content)


def serve_file(path, extension=''):

    if not os.path.exists(path):
        raise ValueError("Path {0} does not exist".format(path))

    hash = md5(path.encode('utf-8')).hexdigest() + extension

    with open(CONFIG) as f:
        config = json.load(f)

    if hash not in config['paths']:

        config['paths'][hash] = os.path.abspath(path)

        with open(CONFIG, 'w') as f:
            json.dump(config, f)

    return '/wwt/' + hash


def load_jupyter_server_extension(nb_server_app):

    web_app = nb_server_app.web_app
    host_pattern = '.*$'

    if not os.path.exists(CONFIG):
        config = {'paths': {}}
        with open(CONFIG, 'w') as f:
            json.dump(config, f)

    mimetypes.add_type('image/fits','.fits')
    mimetypes.add_type('image/fits','.fts')
    mimetypes.add_type('image/fits','.fit')

    route_pattern = url_path_join(web_app.settings['base_url'], '/wwt/(.*)')
    web_app.add_handlers(host_pattern, [(route_pattern, WWTFileHandler)])



from pywwt.qt import WWTQtClient
from astropy import units as u
from astropy.coordinates import concatenate, SkyCoord

# open widget, render at end of each section
wwt = WWTQtClient(size=(600,400), block_until_ready=True)

# big_dipper.png
bd = concatenate((SkyCoord.from_name('Alkaid'), SkyCoord.from_name('Mizar'),
                  SkyCoord.from_name('Alioth'), SkyCoord.from_name('Megrez'),
                  SkyCoord.from_name('Phecda'), SkyCoord.from_name('Merak'),
                  SkyCoord.from_name('Dubhe')))
wwt.center_on_coordinates(SkyCoord.from_name('Megrez'))
       
line = wwt.add_line(bd, width=3 * u.pixel)
wwt.wait()
wwt.render('images/big_dipper.png')

# big_dipper2.png
line.add_point(SkyCoord.from_name('Megrez'))
line.color = 'salmon'
wwt.wait()
wwt.render('images/big_dipper2.png')

# polygons.png
wwt.center_on_coordinates(SkyCoord.from_name('eta orion'))
wwt.crosshairs = False

body = concatenate((SkyCoord.from_name('zeta orion'),SkyCoord.from_name('betelgeuse'),SkyCoord.from_name('bellatrix'),SkyCoord.from_name('delta orion')))
club = concatenate((SkyCoord.from_name('xi orionis'),SkyCoord.from_name('chi2 orionis'),SkyCoord.from_name('chi1 orionis'),SkyCoord.from_name('67 orionis')))
head = concatenate((SkyCoord.from_name('betelgeuse'),SkyCoord.from_name('lambda orion'),SkyCoord.from_name('bellatrix')))
bow = concatenate((SkyCoord.from_name('omicron2 orionis'),SkyCoord.from_name('pi2 orionis'),SkyCoord.from_name('pi3 orionis'),SkyCoord.from_name('pi4 orionis'),SkyCoord.from_name('pi5 orionis')))

a1 = concatenate((SkyCoord.from_name('betelgeuse'),
                  SkyCoord.from_name('xi orionis')))
a2 = concatenate((SkyCoord.from_name('bellatrix'),
                  SkyCoord.from_name('pi3 orionis')))
l1 = concatenate((SkyCoord.from_name('zeta orionis'),
                  SkyCoord.from_name('kappa orionis')))
l2 = concatenate((SkyCoord.from_name('delta orionis'),SkyCoord.from_name('eta orionis'),SkyCoord.from_name('rigel')))
blt = concatenate((SkyCoord.from_name('zeta orionis'),
                    SkyCoord.from_name('delta orionis')))
string = concatenate((SkyCoord.from_name('omicron2 orionis'),
                      SkyCoord.from_name('pi5 orionis')))

orb = wwt.add_polygon(body, fill=True, fill_color='lightslategray', opacity=.4,
                      line_color='lightslategray', line_width=3*u.pixel)
orc = wwt.add_polygon(club, fill=True, fill_color='saddlebrown', opacity=.4,
                      line_color ='saddlebrown', line_width=3*u.pixel)
orh = wwt.add_polygon(head, fill=True, fill_color='rosybrown', opacity=.4,
                      line_color='lightslategray', line_width=2*u.pixel)
orw = wwt.add_polygon(bow, opacity=.4,
                      line_color='saddlebrown', line_width=2*u.pixel)

ora1 = wwt.add_line(a1, color='lightslategray')
ora2 = wwt.add_line(a2, color='lightslategray')
orl1 = wwt.add_line(l1, color='lightslategray', width=2*u.pixel)
orl2 = wwt.add_line(l2, color='lightslategray', width=2*u.pixel)
orblt = wwt.add_line(blt, color='azure', width=8*u.pixel)
orstr = wwt.add_line(string, color='azure')

wwt.wait()
wwt.render('images/polygons.png')

# circles.png
wwt.center_on_coordinates(SkyCoord(190, -55, unit=u.deg))
crc1 = wwt.add_circle(SkyCoord(188, -57, unit=u.deg), radius=10 * u.degree,
                      fill=True, fill_color='#008CA8')
crc2 = wwt.add_circle(radius=10 * u.pixel, opacity=.4,
                      fill=True, fill_color='#C4D600')
wwt.wait()
wwt.render('images/circles.png')

# stgo_view.png
wwt.constellation_boundaries = True
wwt.constellation_figures = True
wwt.constellation_boundary_color = 'azure'
wwt.constellation_figure_color = '#D3BC8D'
wwt.constellation_selection_color = (1, 0, 1)
wwt.local_horizon_mode = True
wwt.location_latitude = -33.4172 * u.deg
wwt.location_longitude = -70.604 * u.deg
wwt.location_altitude = 300 * u.meter
wwt.wait()
wwt.render('images/stgo_view.png')

# dust_on_gamma.png
wwt.clear_annotations()
wwt.constellation_boundaries = False
wwt.constellation_figures = False
wwt.local_horizon_mode = False

wwt.center_on_coordinates(SkyCoord(144.545, -68.5, unit=u.deg))
wwt.background = 'Fermi LAT 8-year (gamma)'
wwt.foreground = 'Planck Dust & Gas'
wwt.foreground_opacity = .75
wwt.wait()
wwt.render('images/dust_on_gamma.png')

# universe.png
wwt.set_view('universe')

wwt.wait()
wwt.render('images/universe.png')

# hdf.png
wwt.set_view('sky')
wwt.clear_annotations()

hdf = SkyCoord(189.206, 62.216, unit=u.deg)
wwt.background = wwt.imagery.visible.sdss
wwt.center_on_coordinates(hdf, fov=2.8*u.arcmin)
fov = wwt.add_fov(wwt.instruments.hst_wfc3_ir, center=hdf,
                  rotate=12*u.deg, line_color='#D4BD8A')
wwt.wait()
wwt.render('images/hdf.png')



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function
import os
from glob import glob
from os.path import join as pjoin


from setupbase import (
    create_cmdclass, install_npm, ensure_targets,
    find_packages, combine_commands, ensure_python,
    get_version, HERE
)

from setuptools import setup


# The name of the project
name = 'pywwt'

# Ensure a valid python version
ensure_python('>=2.7')

# Get our version
version = get_version(pjoin(name, '_version.py'))

nb_path = pjoin(HERE, name, 'nbextension', 'static')
lab_path = pjoin(HERE, name, 'labextension')

# Representative files that should exist after a successful build
jstargets = [
    pjoin(nb_path, 'index.js'),
    pjoin(HERE, 'lib', 'plugin.js'),
]

package_data_spec = {
    name: [
        'nbextension/static/*.*js*',
        'nbextension/static/*.html',
        'labextension/*.tgz',
        'tests/data/*/*.png'
    ]
}

data_files_spec = [
    ('share/jupyter/nbextensions/pywwt', nb_path, '*.js*'),
    ('share/jupyter/nbextensions/pywwt', nb_path, '*.html'),
    ('share/jupyter/lab/extensions', lab_path, '*.tgz'),
    ('etc/jupyter/nbconfig/notebook.d' ,
     os.path.join(HERE, 'jupyter.d', 'notebook.d'), 'pywwt.json'),
    ('etc/jupyter/jupyter_notebook_config.d' ,
     os.path.join(HERE, 'jupyter.d', 'jupyter_notebook_config.d'), 'pywwt.json')]


cmdclass = create_cmdclass('jsdeps', package_data_spec=package_data_spec,
    data_files_spec=data_files_spec)
cmdclass['jsdeps'] = combine_commands(
    install_npm(HERE, build_cmd='build:all'),
    ensure_targets(jstargets),
)

with open('README.rst') as f:
    LONG_DESCRIPTION = f.read()

setup_args = dict(
    name            = name,
    description     = 'WorldWide Telescope from Python',
    long_description = LONG_DESCRIPTION,
    version         = version,
    scripts         = glob(pjoin('scripts', '*')),
    cmdclass        = cmdclass,
    packages        = find_packages(),
    author          = 'Thomas P. Robitaille, O. Justin Otor, and John ZuHone',
    author_email    = 'thomas.robitaille@gmail.com',
    url             = 'https://github.com/WorldWideTelescope/pywwt',
    license         = 'BSD',
    platforms       = "Linux, Mac OS X, Windows",
    keywords        = ['Jupyter', 'Widgets', 'IPython'],
    classifiers     = [
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Topic :: Multimedia :: Graphics',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Framework :: Jupyter',
    ],
    include_package_data = True,
    install_requires = [
        'numpy>=1.9',
        'matplotlib>1.5',
        'astropy>=1.0',
        'requests',
        'beautifulsoup4',
        'python-dateutil',
        'lxml',
        'ipywidgets>=7.0.0',
        'ipyevents',
        'traitlets',
        'reproject>=0.4',
        'qtpy',
        'flask',
        'flask-cors',
        'six'
    ],
    extras_require = {
        'test': [
            'pytest',
            'pytest-cov',
        ],
        'docs': [
            'sphinx>=1.6',
            'sphinx-automodapi',
            'numpydoc',
            'sphinx_rtd_theme',
            'jupyter_sphinx',
        ],
        'qt': [
            'PyQt5',
            'PyQtWebEngine',
        ]
    },
    entry_points = {
    },
)

if __name__ == '__main__':
    setup(**setup_args)



from pywwt.qt import WWTQtClient
from astropy import units as u
from astropy.coordinates import concatenate, SkyCoord

from astropy.table import Table

wwt = WWTQtClient(size=(600,400), block_until_ready=True)

OEC = 'https://worldwidetelescope.github.io/pywwt/data/open_exoplanet_catalogue.csv'
table = Table.read(OEC, delimiter=',', format='ascii.basic')

layer = wwt.layers.add_table_layer(table=table, frame='Sky',
                                  lon_att='ra', lat_att='dec', lon_unit='deg')

layer.size_scale = 50

kepler = SkyCoord(76.532562, 13.289502, unit='deg', frame='galactic')
wwt.center_on_coordinates(kepler, fov=40 * u.deg)

wwt.wait(3)

wwt.render('images/data_layers_kepler.png')

layer.remove()

wwt.set_view('solar system')

wwt.wait(5)

wwt.solar_system.track_object('Earth')

wwt.wait(10)


EARTHQUAKES = 'https://worldwidetelescope.github.io/pywwt/data/earthquakes_2010.csv'
table = Table.read(EARTHQUAKES, delimiter=',', format='ascii.basic')

layer = wwt.layers.add_table_layer(table=table, frame='Earth',
                                  lon_att='longitude', lat_att='latitude')

wwt.wait(3)

wwt.render('images/data_layers_earthquakes.png')

wwt.wait()

# layer.lon_att = 'longitude'
#
# layer.alt_att = 'depth'
